/*
    Copyright 2014 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt

VC\bin\vcvars32.bat
VC\bin\cl.exe steamuserip.cpp /I ..\sdk\public ..\sdk\redistributable_bin\steam_api.lib /Ox
*/
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <windows.h>
#pragma comment(lib, "ws2_32.lib")

#define STEAM_API_NODLL
#include "steam/steam_api.h"



#define VER     "0.1a"



CSteamID    m_SteamIDLocalUser;



void show_steamid_info(char *info, CSteamID id) {
    printf("%s %llu - %u %u %u %u\n", info,
        id.ConvertToUint64(),
        id.GetAccountID(), id.GetUnAccountInstance(), id.GetEAccountType(), id.GetEUniverse());
}        



void get_ip_address(CSteamID targetID, int resolve_ip) {
    FriendGameInfo_t    fgi;
    struct in_addr  addr;
    struct hostent  *h;
    int     ip,
            channel;
    uint32  *ips = NULL;
    int     i,
            n;
    char    *p;

    if(!targetID.ConvertToUint64()) return;
    if(targetID == m_SteamIDLocalUser) return;

    printf("\n");
    show_steamid_info("  target ID", targetID);
    //SteamFriends()->RequestUserInformation(targetID, TRUE);
    printf("  Target Name %s\n", SteamFriends()->GetFriendPersonaName(targetID));    
    p = (char *)SteamFriends()->GetPlayerNickname(targetID);
    if(p) printf("  Nick Name %s\n", p);

    if(SteamFriends()->GetFriendGamePlayed(targetID, &fgi)) {
        //show_steamid_info("m_gameID", fgi.m_gameID);
        printf("  %08x %d %d\n", fgi.m_unGameIP, fgi.m_usGamePort, fgi.m_usQueryPort);
        show_steamid_info("  m_steamIDLobby", fgi.m_steamIDLobby);
    }
	switch(SteamFriends()->GetFriendPersonaState(targetID)) {
        k_EPersonaStateOffline: p = "friend is not currently logged on"; break;
        k_EPersonaStateOnline: p = "friend is logged on"; break;
        k_EPersonaStateBusy: p = "user is on, but busy"; break;
        k_EPersonaStateAway: p = "auto-away feature"; break;
        k_EPersonaStateSnooze: p = "auto-away for a long time"; break;
        k_EPersonaStateLookingToTrade: p = "Online, trading"; break;
        k_EPersonaStateLookingToPlay: p = "Online, wanting to play"; break;
        default: p = NULL; break;
    }
    if(p) printf("  GetFriendPersonaState %s\n", p);

    //SteamFriends()->RequestFriendRichPresence(targetID);
    n = SteamFriends()->GetFriendRichPresenceKeyCount(targetID);
    for(i = 0; i < n; i++) {
        p = (char *)SteamFriends()->GetFriendRichPresenceKeyByIndex(targetID, i);
        printf("  %s: %s\n", p, (char *)SteamFriends()->GetFriendRichPresence(targetID, p));
    }


    channel = 0;
    printf("\n- retrieving IP addresses\n");
    for(ip = 0;; ip++) {
        SteamNetworking()->AllowP2PPacketRelay(FALSE);  // FALSE or TRUE is probably the same, let's use TRUE
        P2PSessionState_t   p2ps;

        //SteamNetworking()->CloseP2PChannelWithUser(targetID, 0);
        SteamNetworking()->CloseP2PSessionWithUser(targetID);   // just to be sure
        for(i = 0;;) {
            channel = i;// & 1;

            // send packets here in the cycle, otherwise after a successful retrieving it may not work for some time
            //SteamNetworking()->SendP2PPacket(targetID, "", 0, k_EP2PSendUnreliableNoDelay, channel);
            //SteamNetworking()->SendP2PPacket(targetID, "", 0, k_EP2PSendReliable, channel);
            SteamNetworking()->SendP2PPacket(targetID, "", 0, k_EP2PSendReliableWithBuffering, channel);

            memset(&p2ps, 0, sizeof(p2ps));
            if(SteamNetworking()->GetP2PSessionState(targetID, &p2ps)) {
                if(p2ps.m_nRemoteIP) break;
            }
            Sleep(50);
            i += 50;
            if(i >= (5 * 1000)) { // SendP2PPacket can take even 20 seconds
                if(!ip) printf("\nError: timeout, this game or the current status of the target don't allow to retrieve his IP address\n");
                goto done;
            }
        }
        //SteamNetworking()->CloseP2PChannelWithUser(targetID, 0);
        SteamNetworking()->CloseP2PSessionWithUser(targetID);   // just to be sure

        for(i = 0; i < ip; i++) {
            if(ips[i] == p2ps.m_nRemoteIP) goto done;
        }

        /*
        printf("  m_bConnectionActive     %d\n", p2ps.m_bConnectionActive);
        printf("  m_bConnecting           %d\n", p2ps.m_bConnecting);
        printf("  m_eP2PSessionError      %d\n", p2ps.m_eP2PSessionError);
        printf("  m_bUsingRelay           %d\n", p2ps.m_bUsingRelay);
        printf("  m_nBytesQueuedForSend   %d\n", p2ps.m_nBytesQueuedForSend);
        printf("  m_nPacketsQueuedForSend %d\n", p2ps.m_nPacketsQueuedForSend);
        printf("  m_nRemoteIP             %s\n", inet_ntoa(addr));
        printf("  m_nRemotePort           %d\n", p2ps.m_nRemotePort);
        */
        addr.s_addr = htonl(p2ps.m_nRemoteIP);
        // p2ps.m_bUsingRelay is ever zero
        printf("\n* %s\n", inet_ntoa(addr));

        ips = (uint32 *)realloc(ips, (ip + 1) * sizeof(* ips));
        ips[ip] = p2ps.m_nRemoteIP;

        //Sleep(500); // for CloseP2P?
    }

done:
    SteamNetworking()->CloseP2PSessionWithUser(targetID);

    if(resolve_ip && ip) {
        printf("\n- resolving to hostnames:\n");
        for(i = 0; i < ip; i++) {
            addr.s_addr = htonl(ips[i]);
            n = addr.s_addr & 0xff;
            if((n == 192) || (n == 10)) continue;   // LAN IP

            h = gethostbyaddr((const char *)&addr, 4, AF_INET);
            if(h) {
                printf("  %s\n", inet_ntoa(addr));
                printf("                %s\n", h->h_name);
                char **pAlias;
                for (pAlias = h->h_aliases; *pAlias; pAlias++) {
                    printf("                %s\n", *pAlias);
                }
            }
        }
    }
    if(ips) free(ips);
}



/*
class CLobbyJoin
{
public:
    int Refresh(CSteamID lobbyID, bool check_if_terminated) {
        if(check_if_terminated) {
            if((time(NULL) - timestamp) > 10) return 1;   // timeout
            return done;
        }
        
        timestamp = time(NULL);
        done = 0;
        joinlobbyID = lobbyID;
		m_CLobbyJoin.Set(
            SteamMatchmaking()->JoinLobby(lobbyID),
            this, &CLobbyJoin::OnLobbyJoinCallback );
        return 0;
    }
private:
    CSteamID joinlobbyID;
    int     done;
    time_t  timestamp;
	CCallResult<CLobbyJoin, LobbyEnter_t> m_CLobbyJoin;
    void OnLobbyJoinCallback( LobbyEnter_t *pCallback, bool bIOFailure ) {
        //if(pCallback->m_ulSteamIDLobby == joinlobbyID.ConvertToUint64()) {
            if(pCallback->m_EChatRoomEnterResponse == k_EChatRoomEnterResponseSuccess) done = 1;
            done = -1;
        //}
    }
};
*/



void scan_lobby(CSteamID lobbyID) {
    CSteamID    targetID;
    int     i,
            n,
            timeout;

    SteamMatchmaking()->JoinLobby(lobbyID);
    Sleep(1000);    // give some time to join

    /* doesn't seem to work... blah, no time to waste
    CLobbyJoin *m_pLobbyJoin = new CLobbyJoin();
    m_pLobbyJoin->Refresh(lobbyID, 0);
    while(!m_pLobbyJoin->Refresh(lobbyID, 1)) {
        SteamAPI_RunCallbacks();
        Sleep(10);
    }
    delete m_pLobbyJoin;
    */

    timeout = 0;
    n = SteamMatchmaking()->GetNumLobbyMembers(lobbyID);
    printf("\n- %d players (including you)\n", n);
    for(i = 0; i < n; i++) {
        targetID = SteamMatchmaking()->GetLobbyMemberByIndex(lobbyID, i);

        // this happens when we are not in the lobby yet
        if(!targetID.ConvertToUint64()) {
            Sleep(100);
            timeout += 100;
            if(timeout >= 3000) break;
            i--;
            continue;
        }

        get_ip_address(targetID, 0);
    }

    SteamMatchmaking()->LeaveLobby(lobbyID);
}



void show_appid_name(AppId_t appID) {
    char    name[1024],
            *p;

    if(SteamAppList()->GetAppName(appID, name, sizeof(name)) < 0)
    if(SteamAppList()->GetAppInstallDir(appID, name, sizeof(name)) < 0)
    if(SteamApps()->GetAppInstallDir(appID, name, sizeof(name)) <= 0)
        name[0] = 0;
    p = strrchr(name, '\\');
    if(!p) p = strrchr(name, '/');
    if(!p) p = name;
    else p++;
    printf("\n### GAME %u - %s ###\n", (int)appID, p);
}



class CLobbyBrowser
{
public:
    int Refresh(bool check_if_terminated) {
        if(check_if_terminated) {
            if((time(NULL) - timestamp) > 10) return 1;   // timeout
            return done;
        }
        
        timestamp = time(NULL);
        done = 0;
		SteamAPICall_t hSteamAPICall = SteamMatchmaking()->RequestLobbyList();
		m_SteamCallResultLobbyMatchList.Set( hSteamAPICall, this, &CLobbyBrowser::OnLobbyMatchListCallback );
        return 0;
    }
private:
    bool    done;
    time_t  timestamp;
	CCallResult<CLobbyBrowser, LobbyMatchList_t> m_SteamCallResultLobbyMatchList;
    void OnLobbyMatchListCallback( LobbyMatchList_t *pCallback, bool bIOFailure ) {
        CSteamID steamIDLobby;
        int     datas;
        char    Key[1024],      // bigger than usual
                Value[8192];    // bigger than usual

        printf("%d lobb%s\n", pCallback->m_nLobbiesMatching, (pCallback->m_nLobbiesMatching == 1) ? "y" : "ies");
        for ( int iLobby = 0; iLobby < pCallback->m_nLobbiesMatching; iLobby++ )
        {
            steamIDLobby = SteamMatchmaking()->GetLobbyByIndex( iLobby );
            printf("\n");
            printf("-----------------------------------------------\n");
            show_steamid_info("lobby", steamIDLobby);
            printf("-----------------------------------------------\n");
            int cLobbyMembers = SteamMatchmaking()->GetNumLobbyMembers( steamIDLobby );
            datas = SteamMatchmaking()->GetLobbyDataCount(steamIDLobby);
            for(int iLobbyData = 0; iLobbyData < datas; iLobbyData++) {
                memset(Key, 0, sizeof(Key));        // not needed
                memset(Value, 0, sizeof(Value));    // not needed
                SteamMatchmaking()->GetLobbyDataByIndex(steamIDLobby, iLobbyData, Key, sizeof(Key), Value, sizeof(Value));
                printf("  %s: %s\n", Key, Value);
            }

            scan_lobby(steamIDLobby);
        }
        done = 1;
    }
};



int main(int argc, char *argv[]) {
    CSteamID    targetID;
    AppId_t     appID;
    uint64      n64;
    char        *id;

    fprintf(stderr,
        "\n"
        "steamuserip %s\n"
        "by Luigi Auriemma\n"
        "e-mail: me@aluigi.org\n"
        "web:    aluigi.org\n"
        "\n", VER);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <appID> <steamID/64/\"\">\n"
            "\n"
            "  If the steamID is 0 or \"\", the tool will get the online lobbies, join them\n"
            "  and scan all the users currently in them (in some games it may work).\n"
            "\n"
            "  The user must be online and playing with the same game of appID.\n"
            "\n"
            "  If you get only a LAN IP address, retry again.\n"
            "  You may get also the Steam relay servers like 146.66.15*.*\n"
            "  Other Steam servers may be 208.78.16*.*\n"
            "  If you get only LAN and relay IP address, retry again.\n"
            "\n"
            "  Remember that the results are game depend, this is not an universal solution.\n"
            "  This tool is only a proof-of-concept, it may be updated in future.\n"
            "\n", argv[0]);
        exit(1);
    }

    SetEnvironmentVariable("SteamAppId", argv[1]);
    SetEnvironmentVariable("SteamGameId", argv[1]);
    appID = atoi(argv[1]);

    id = argv[2];
    if(strstr(id, "://")) {
        if(strstr(id, "/profiles/")) {
            n64 = _atoi64(strrchr(id, '/') + 1);
        } else {
            printf("\nError: add /?xml=1 to the URL to retrieve the SteamID64 value\n");
            exit(1);
        }
    } else if(strchr(id, ':')) {
        n64 = 0x0110000100000000 | (_atoi64(strrchr(id, ':') + 1) << 1) | atoi(strchr(id, ':') + 1);
    } else {
        n64 = _atoi64(id);
    }
    targetID.SetFromUint64(n64);

    if(!SteamAPI_Init()) {
        printf("\nError: Steam is not running\n");
        exit(1);
    }
    if(!SteamUser()->BLoggedOn()) printf("Steam user is not logged in\n");
    m_SteamIDLocalUser = SteamUser()->GetSteamID();
    show_steamid_info("Your ID", m_SteamIDLocalUser);
    printf("Your Name %s\n", SteamFriends()->GetPersonaName());


    show_appid_name(appID);

    // IP addresses
    if(!targetID.ConvertToUint64()) {
        printf("- request list of lobbies:\n");    
        CLobbyBrowser *m_pLobbyBrowser = new CLobbyBrowser();
        m_pLobbyBrowser->Refresh(0);
        while(!m_pLobbyBrowser->Refresh(1)) {
            SteamAPI_RunCallbacks();
            Sleep(10);
        }

    } else {
        get_ip_address(targetID, 0);
    }

    printf("\n- done\n");
    SteamAPI_Shutdown();    
    return 0;
}
